# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Family of exceptions for package API.

    * ``Omniexception``: Base for all package exceptions
    * ``Omnierror``: Base for all package errors
'''


from . import __


class Omniexception( BaseException, metaclass = __.ImmutableClass ):
    ''' Base for all exceptions raised by package API. '''
    # TODO: Class attribute concealment.
    # TODO: Instance attribute concealment and immutability.

    _attribute_visibility_includes_: __.cabc.Collection[ str ] = (
        frozenset( ( '__cause__', '__context__', ) ) )


class Omnierror( Omniexception, Exception ):
    ''' Base for error exceptions raised by package API. '''


class ContentAcquireFailure( Omnierror ):
    ''' Failure to acquire content from location. '''

    def __init__( self, location: str | __.Path ):
        super( ).__init__( f"Could not acquire content from '{location}'." )


class ContentDecodeFailure( Omnierror ):
    ''' Failure to decode content as character set from location. '''

    def __init__( self, location: str | __.Path, charset: str ):
        super( ).__init__(
            f"Could not decode content at '{location}' "
            f"as character set '{charset}'." )


class ContentUpdateFailure( Omnierror ):
    ''' Failure to update content at location. '''

    def __init__( self, location: str | __.Path ):
        super( ).__init__( f"Could not update content at '{location}'." )


class DifferencesProcessFailure( Omnierror ):
    ''' Failure during diff processing. '''

    def __init__( self, reason: str ):
        super( ).__init__( f"Could not process changes. Reason: {reason}" )


class EditorFailure( Omnierror ):
    ''' Failure while operating editor. '''

    def __init__( self, cause: str | Exception ):
        super( ).__init__( f"Could not edit content. Cause: {cause}" )


class LocationInvalidity( Omnierror ):
    ''' Invalid location. '''

    def __init__( self, location: str | __.Path ):
        super( ).__init__( f"Invalid location '{location}'." )


class MimeogramFormatEmpty( Omnierror ):
    ''' Attempt to format empty mimeogram. '''

    def __init__( self ):
        super( ).__init__( "Cannot format empty mimeogram." )


class MimeogramParseFailure( Omnierror ):
    ''' Failure to parse mimeogram content. '''

    def __init__( self, reason: str ):
        super( ).__init__( f"Could not parse mimeogram. Reason: {reason}" )


class PagerFailure( Omnierror ):
    ''' Failure while operating pager. '''

    def __init__( self, cause: str | Exception ):
        super( ).__init__( f"Could not display content. Cause: {cause}" )


class ProgramAbsenceError( Omnierror ):
    ''' Could not discover valid editor. '''

    def __init__( self, species: str ):
        super( ).__init__( f"Could not discover valid {species}." )


class TextualMimetypeInvalidity( Omnierror ):
    ''' Invalid textual MIME type for content at location. '''

    def __init__( self, location: str | __.Path, mimetype: str ):
        super( ).__init__(
            f"Invalid MIME type '{mimetype}' for content at '{location}'." )


class UrlSchemeNoSupport( Omnierror ):
    ''' Unsupported URL scheme. '''

    def __init__( self, url: str ):
        super( ).__init__( f"Scheme of URL '{url}' is not supported." )


class UserOperateCancellation( Omniexception ):
    ''' Operation cancelled by user. '''

    def __init__( self, cause: BaseException ):
        super( ).__init__( f"Operation cancelled by user. Cause: {cause}" )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' System editor interaction. '''


from __future__ import annotations

from . import __


_scribe = __.produce_scribe( __name__ )


def discover_editor( ) -> __.cabc.Callable[ [ str ], str ]:
    ''' Discovers editor and returns executor function. '''
    from shutil import which
    from subprocess import run # nosec B404
    editor = __.os.environ.get( 'VISUAL' ) or  __.os.environ.get( 'EDITOR' )
    for editor_ in filter(
        None,
        # Editors, ranked by "friendliness", not by personal preference.
        ( editor, 'code', 'nano', 'emacs', 'nvim', 'vim' )
    ):
        if ( editor := which( editor_ ) ): break
    else: editor = ''
    match editor:
        case 'code': posargs = ( '--wait', )
        case _: posargs = ( )

    if editor:

        # TODO? async
        def editor_executor( filename: str ) -> str:
            ''' Executes editor with file. '''
            run( ( editor, *posargs, filename ), check = True ) # nosec B603
            with open( filename, 'r', encoding = 'utf-8' ) as stream:
                return stream.read( )

        return editor_executor

    _scribe.error(
        "No suitable text editor found. "
        "Please install a console-based editor "
        "or set the 'EDITOR' environment variable to your preferred editor." )
    # TODO: Add suggestions.
    from .exceptions import ProgramAbsenceError
    raise ProgramAbsenceError( 'editor' )


def edit_content( # pylint: disable=too-many-locals
    content: str = '', *,
    suffix: str = '.md',
    editor_discoverer: __.cabc.Callable[
        [ ], __.cabc.Callable[ [ str ], str ] ] = discover_editor,
) -> str:
    ''' Edits content via discovered editor. '''
    from .exceptions import EditorFailure, ProgramAbsenceError
    try: editor = editor_discoverer( )
    except ProgramAbsenceError: return content
    import tempfile
    from pathlib import Path
    # Using delete = False to handle file cleanup manually. This ensures
    # the file handle is properly closed before the editor attempts to read it,
    # which is particularly important on Windows where open files cannot be
    # simultaneously accessed by other processes without a read share.
    with tempfile.NamedTemporaryFile(
        mode = 'w', suffix = suffix, delete = False, encoding = 'utf-8'
    ) as tmp:
        filename = tmp.name
        tmp.write( content )
    try: return editor( filename ) # noqa: TRY101
    except Exception as exc: raise EditorFailure( cause = exc ) from exc
    finally:
        try: Path( filename ).unlink( )
        except Exception: # pylint: disable=broad-exception-caught
            _scribe.exception( f"Failed to cleanup {filename}" )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Content differences management. '''


from __future__ import annotations

from . import __
from . import interfaces as _interfaces
from . import parts as _parts


_scribe = __.produce_scribe( __name__ )


class ConsoleDisplay(
    _interfaces.DifferencesDisplay, decorators = ( __.standard_dataclass, )
):
    ''' Default display of differences to console. '''

    async def __call__( self, lines: __.cabc.Sequence[ str ] ) -> None:
        from .display import display_content
        if self.inline_threshold >= len( lines ):
            for line in lines: print( line )
            return
        diff = '\n'.join( lines )
        display_content( diff, suffix = '.diff' )


class ConsoleInteractor( _interfaces.DifferencesInteractor ):
    ''' Default console-based interaction handler. '''

    async def __call__(
        self,
        lines: __.cabc.Sequence[ str ],
        display: _interfaces.DifferencesDisplay
    ) -> bool:
        # TODO: Display hunk number.
        from readchar import readkey
        await display( lines )
        menu = "Apply this change? (y)es, (n)o, (v)iew"
        while True:
            print( f"\n{menu} > ", end = '' )
            try: choice = readkey( ).lower( )
            except ( EOFError, KeyboardInterrupt ):
                print( ) # Add newline to avoid output mangling
                return False
            print( choice ) # Echo.
            match choice:
                case 'y': return True
                case 'n': return False
                case 'v': await display( lines )
                case _:
                    if choice.isprintable( ):
                        print( f"Invalid choice: {choice}" )
                    else: print( "Invalid choice." )


async def select_segments(
    target: _parts.Target, revision: str,
    display: __.Absential[ _interfaces.DifferencesDisplay ] = __.absent,
    interactor: __.Absential[ _interfaces.DifferencesInteractor ] = __.absent,
) -> str:
    ''' Selects which diff hunks to apply. '''
    # TODO: Use global state for instance configuration.
    if __.is_absent( display ): display = ConsoleDisplay( )
    if __.is_absent( interactor ): interactor = ConsoleInteractor( )
    # TODO: Acquire destination content from cache.
    part = target.part
    original = (
        await __.acquire_text_file_async(
            target.destination, charset = part.charset ) )
    original = part.linesep.normalize( original )
    if original == revision:
        print( "No changes" )
        return revision
    try:
        revision_ = (
            await _select_segments(
                original, revision,
                display = display, interactor = interactor ) )
    except Exception: # pylint: disable=broad-exception-caught
        _scribe.exception( "Could not process changes" )
        return revision
    return revision_


def _format_segment( # pylint: disable=too-many-arguments,too-many-locals
    current_lines: list[ str ],
    revision_lines: list[ str ],
    i1: int, i2: int,
    j1: int, j2: int,
    context_lines: int = 3,
) -> list[ str ]:
    ''' Formats change block with context lines. '''
    # Calculate context ranges with bounds checking
    start = max( 0, i1 - context_lines )
    end = min( len( current_lines ), i2 + context_lines )
    # Build diff display
    # TODO? Convert non-printables into printable sequences.
    diff: list[ str ] = [ ]
    diff.append(
        f"@@ -{i1 + 1},{i2 - i1} +{j1 + 1},{j2 - j1} @@" )
    for idx in range( start, i1 ):
        diff.append( f" {current_lines[ idx ]}" )
    for idx in range( i1, i2 ):
        diff.append( f"-{current_lines[ idx ]}" )
    for idx in range( j1, j2 ):
        diff.append( f"+{revision_lines[ idx ]}" )
    for idx in range( i2, end ):
        diff.append( f" {current_lines[ idx ]}" )
    return diff


async def _select_segments( # pylint: disable=too-many-locals
    current: str,
    revision: str,
    display: _interfaces.DifferencesDisplay,
    interactor: _interfaces.DifferencesInteractor,
) -> str:
    from patiencediff import PatienceSequenceMatcher # pyright: ignore
    current_lines = current.split( '\n' )
    revision_lines = revision.split( '\n' )
    matcher = PatienceSequenceMatcher( # pyright: ignore
        None, current_lines, revision_lines )
    result: list[ str ] = [ ]
    for op, i1, i2, j1, j2 in matcher.get_opcodes( ):
        if op == 'equal': # pylint: disable=magic-value-comparison
            result.extend( current_lines[ i1:i2 ] )
            continue
        diff_lines = _format_segment(
            current_lines, revision_lines,
            i1, i2, j1, j2,
            context_lines = display.context_lines )
        if not await interactor( diff_lines, display ):
            result.extend( current_lines[ i1:i2 ] )
            continue
        result.extend( revision_lines[ j1:j2 ] )
    return '\n'.join( result )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Core entities for filesystem protection. '''


from __future__ import annotations

from .. import __


class Protector( # pylint: disable=invalid-metaclass
    __.typx.Protocol,
    metaclass = __.ImmutableStandardProtocolDataclass,
    decorators = ( __.standard_dataclass, __.typx.runtime_checkable ),
):
    ''' Filesystem protection checker. '''

    @__.abc.abstractmethod
    def verify( self, path: __.Path ) -> Status:
        ''' Verifies if a path should be protected. '''
        raise NotImplementedError


class Status(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Protection status for location. '''

    path: __.Path
    reason: __.typx.Optional[ Reasons ] = None
    active: bool = False

    def __bool__( self ): return self.active

    @property
    def description( self ) -> str:
        ''' Human-readable description of protection. '''
        if not self.active: return 'Not protected'
        return (
            f"Protected: {self.reason.value}"
            if self.reason else 'Protected' )


class Reasons( __.enum.Enum ):
    ''' Reasons why location may be protected. '''

    Concealment =       'Hidden file or directory'
    Credentials =       'Credentials or secrets location'
    CustomAddition =    'User-specified custom location'
    OsDirectory =       'Operating system directory'
    PlatformSensitive = 'Platform-sensitive location'
    UserConfiguration = 'User configuration directory'
    VersionControl =    'Version control internals'



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Exchange of MIME bundles with LLMs. '''


from . import __
from . import acquirers
from . import apply
from . import cli
from . import create
from . import differences
from . import display
from . import edit
from . import exceptions
from . import formatters
from . import interactions
from . import parsers
from . import updaters

# TODO: Export various module contents.


__version__ = '1.0a7'


def main( ):
    ''' Entrypoint. '''
    cli.execute( )


__.reclassify_modules_as_immutable( __name__ )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Sensitive filesystem locations on Unix/POSIX/Linux. '''


from __future__ import annotations

from . import __


_scribe = __.produce_scribe( __name__ )


def discover_system_paths( ) -> frozenset[ __.Path ]:
    ''' Discovers Unix and other relevant system paths. '''
    paths: set[ __.Path ] = set( )
    paths.update( map(
        __.Path,
        (   '/bin', '/boot', '/dev', '/etc', '/opt',
            '/proc', '/sbin', '/sys', '/usr',
        ) ) )
    #if _detect_wsl( ): paths.update( _discover_wsl_system_paths( ) )
    return frozenset( paths )


# def _detect_wsl( ) -> bool:
#     ''' Checks if running in WSL environment. '''
#     # TODO: If environment vars are not set, then return False.
#     #       Else, proceed to more expensive check to ensure sane WSL.
#     wsl_vars = ( 'WSL_DISTRO_NAME', 'WSL_INTEROP' )
#     for var in wsl_vars:
#         if __.os.environ.get( var ):
#             _scribe.debug( f"WSL detected via {var} environment variable" )
#             return True
#     try:
#         with open( '/proc/version', 'r' ) as f: # pylint: disable=unspecified-encoding
#             version_info = f.read( ).lower( )
#     except Exception as exc: # pylint: disable=broad-exception-caught
#         _scribe.debug( f"Error checking /proc/version: {exc}" )
#     if 'microsoft' in version_info: # pylint: disable=magic-value-comparison
#         _scribe.debug( "WSL detected via /proc/version" )
#         return True
#     if 'microsoft' in __.os.uname( ).release.lower( ): # pylint: disable=magic-value-comparison
#         _scribe.debug( "WSL detected via kernel release string." )
#         return True
#     return False


# def _discover_wsl_system_paths( ) -> set[ __.Path ]:
#     # TODO: Better method to determine Windows syspaths from WSL.
#     #       Possibly like https://stackoverflow.com/a/55635008/14833542
#     #       since the proper ones cannot be guaranteed via 'WSLENV'.
#     #       But, running 'cmd' is expensive.
#     from . import windows
#     paths = windows.discover_system_paths( )
#     paths_: set[ __.Path ] = set( )
#     for path in paths:
#         parts = path.parts
#         if 1 >= len( parts ): continue
#         drive = parts[ 0 ][ 0 ].lower( ) # TODO? Consider UNC paths.
#         paths_.add( __.Path( f"/mnt/{drive}" ).joinpath( *parts[ 1 : ] ) )
#     _scribe.debug( "Calculated {} WSL system paths.".format( len( paths_ ) ) )
#     return paths_



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Content acquisition from various sources. '''


from __future__ import annotations

import aiofiles as _aiofiles
import httpx as _httpx

from . import __
from . import exceptions as _exceptions
from . import parts as _parts


_scribe = __.produce_scribe( __name__ )


async def acquire( # pylint: disable=too-many-locals
    auxdata: __.Globals, sources: __.cabc.Sequence[ str | __.Path ]
) -> __.cabc.Sequence[ _parts.Part ]:
    ''' Acquires content from multiple sources. '''
    from urllib.parse import urlparse
    options = auxdata.configuration.get( 'acquire-parts', { } )
    strict = options.get( 'fail-on-invalid', False )
    recursive = options.get( 'recurse-directories', False )
    tasks: list[ __.cabc.Coroutine[ None, None, _parts.Part ] ] = [ ]
    for source in sources:
        path = __.Path( source )
        url_parts = (
            urlparse( source ) if isinstance( source, str )
            else urlparse( str( source ) ) )
        scheme = 'file' if path.drive else url_parts.scheme
        match scheme:
            case '' | 'file':
                tasks.extend( _produce_fs_tasks( source, recursive ) )
            case 'http' | 'https':
                tasks.append( _produce_http_task( str( source ) ) )
            case _:
                raise _exceptions.UrlSchemeNoSupport( str( source ) )
    if strict: return await __.gather_async( *tasks )
    results = await __.gather_async( *tasks, return_exceptions = True )
    # TODO: Factor into '__.generics.extract_results_filter_errors'.
    values: list[ _parts.Part ] = [ ]
    for result in results:
        if result.is_error( ):
            _scribe.warning( str( result.error ) )
            continue
        values.append( result.extract( ) )
    return tuple( values )


async def _acquire_from_file( location: __.Path ) -> _parts.Part:
    ''' Acquires content from text file. '''
    from .exceptions import ContentAcquireFailure, ContentDecodeFailure
    try:
        async with _aiofiles.open( location, 'rb' ) as f:
            content_bytes = await f.read( )
    except Exception as exc: raise ContentAcquireFailure( location ) from exc
    mimetype, charset = _detect_mimetype_and_charset( content_bytes, location )
    if charset is None: raise ContentDecodeFailure( location, '???' )
    linesep = _parts.LineSeparators.detect_bytes( content_bytes )
    # TODO? Separate error for newline issues.
    if linesep is None: raise ContentDecodeFailure( location, charset )
    try: content = content_bytes.decode( charset )
    except Exception as exc:
        raise ContentDecodeFailure( location, charset ) from exc
    _scribe.debug( f"Read file: {location}" )
    return _parts.Part(
        location = str( location ),
        mimetype = mimetype,
        charset = charset,
        linesep = linesep,
        content = linesep.normalize( content ) )


async def _acquire_via_http( # pylint: disable=too-many-locals
    client: _httpx.AsyncClient, url: str
) -> _parts.Part:
    ''' Acquires content via HTTP/HTTPS. '''
    from .exceptions import ContentAcquireFailure, ContentDecodeFailure
    try:
        response = await client.get( url )
        response.raise_for_status( )
    except Exception as exc: raise ContentAcquireFailure( url ) from exc
    mimetype = (
        response.headers.get( 'content-type', 'application/octet-stream' )
        .split( ';' )[ 0 ].strip( ) )
    content_bytes = response.content
    charset = response.encoding or _detect_charset( content_bytes )
    if charset is None: raise ContentDecodeFailure( url, '???' )
    if not _is_textual_mimetype( mimetype ):
        mimetype, _ = (
            _detect_mimetype_and_charset(
                content_bytes, url, charset = charset ) )
    linesep = _parts.LineSeparators.detect_bytes( content_bytes )
    # TODO? Separate error for newline issues.
    if linesep is None: raise ContentDecodeFailure( url, charset )
    try: content = content_bytes.decode( charset )
    except Exception as exc:
        raise ContentDecodeFailure( url, charset ) from exc
    _scribe.debug( f"Fetched URL: {url}" )
    return _parts.Part(
        location = url,
        mimetype = mimetype,
        charset = charset,
        linesep = linesep,
        content = linesep.normalize( content ) )


# VCS directories to skip during traversal
_VCS_DIRS = frozenset( ( '.git', '.svn', '.hg', '.bzr' ) )
def _collect_directory_files(
    directory: __.Path, recursive: bool
) -> list[ __.Path ]:
    ''' Collects and filters files from directory hierarchy. '''
    import gitignorefile
    cache = gitignorefile.Cache( )
    paths: list[ __.Path ] = [ ]
    _scribe.debug( f"Collecting files in directory: {directory}" )
    for entry in directory.iterdir( ):
        if entry.is_dir( ) and entry.name in _VCS_DIRS:
            _scribe.debug( f"Ignoring VCS directory: {entry}" )
            continue
        if cache( str( entry ) ):
            _scribe.debug( f"Ignoring path (matched by .gitignore): {entry}" )
            continue
        if entry.is_dir( ) and recursive:
            paths.extend( _collect_directory_files( entry, recursive ) )
        elif entry.is_file( ): paths.append( entry )
    return paths


def _detect_charset( content: bytes ) -> str | None:
    # TODO: Pyright bug: `None is charset` != `charset is None`
    from chardet import detect
    charset = detect( content )[ 'encoding' ]
    if charset is None: return charset
    if charset.startswith( 'utf' ): return charset
    match charset:
        case 'ascii': return 'utf-8' # Assume superset.
        case _: pass
    # Shake out false positives, like 'MacRoman'.
    try: content.decode( 'utf-8' )
    except UnicodeDecodeError: return charset
    return 'utf-8'


def _detect_mimetype( content: bytes, location: str | __.Path ) -> str | None:
    from mimetypes import guess_type
    from puremagic import PureError, from_string # pyright: ignore
    try: return from_string( content, mime = True )
    except ( PureError, ValueError ):
        return guess_type( str( location ) )[ 0 ]


def _detect_mimetype_and_charset(
    content: bytes,
    location: str | __.Path, *,
    mimetype: __.Absential[ str ] = __.absent,
    charset: __.Absential[ str ] = __.absent,
) -> tuple[ str, str | None ]:
    from .exceptions import TextualMimetypeInvalidity
    if __.is_absent( mimetype ):
        mimetype_ = _detect_mimetype( content, location )
    else: mimetype_ = mimetype
    if __.is_absent( charset ):
        charset_ = _detect_charset( content )
    else: charset_ = charset
    if not mimetype_:
        if charset_: mimetype_ = 'text/plain' # pylint: disable=redefined-variable-type
        else: mimetype_ = 'application/octet-stream'
    if not _is_textual_mimetype( mimetype_ ):
        raise TextualMimetypeInvalidity( location, mimetype_ )
    return mimetype_, charset_


# MIME types that are considered textual beyond those starting with 'text/'
_TEXTUAL_MIME_TYPES = frozenset( (
    'application/json',
    'application/xml',
    'application/xhtml+xml',
    'application/javascript',
    'image/svg+xml',
) )
def _is_textual_mimetype( mimetype: str ) -> bool:
    ''' Checks if MIME type represents textual content. '''
    _scribe.debug( f"MIME type: {mimetype}" )
    if mimetype.startswith( ( 'text/', 'application/x-', 'text/x-' ) ):
        return True
    return mimetype in _TEXTUAL_MIME_TYPES


def _produce_fs_tasks(
    location: str | __.Path, recursive: bool = False
) -> tuple[ __.cabc.Coroutine[ None, None, _parts.Part ], ...]:
    location_ = __.Path( location )
    if location_.is_file( ) or location_.is_symlink( ):
        return ( _acquire_from_file( location_ ), )
    if location_.is_dir( ):
        files = _collect_directory_files( location_, recursive )
        return tuple( _acquire_from_file( f ) for f in files )
    raise _exceptions.ContentAcquireFailure( location )


def _produce_http_task(
    url: str
) -> __.cabc.Coroutine[ None, None, _parts.Part ]:
    # TODO: URL object rather than string.
    # TODO: Reuse clients for common hosts.

    async def _execute_session( ) -> _parts.Part:
        async with _httpx.AsyncClient( # nosec B113
            follow_redirects = True
        ) as client: return await _acquire_via_http( client, url )

    return _execute_session( )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Creation of mimeograms. '''
# TODO? Use BSD sysexits.


from __future__ import annotations

from . import __
from . import interfaces as _interfaces


_scribe = __.produce_scribe( __name__ )


class Command(
    _interfaces.CliCommand,
    decorators = ( __.standard_dataclass, __.standard_tyro_class ),
):
    ''' Creates mimeogram from filesystem locations or URLs. '''

    sources: __.typx.Annotated[
        __.tyro.conf.Positional[ list[ str ] ],
        __.tyro.conf.arg(
            help = "Filesystem locations or URLs.",
            prefix_name = False ),
    ]
    clip: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.tyro.conf.arg(
            aliases = ( '--clipboard', '--to-clipboard' ),
            help = "Copy mimeogram to clipboard." ),
    ] = None
    edit: __.typx.Annotated[
        bool,
        __.tyro.conf.arg(
            aliases = ( '-e', '--edit-message' ),
            help = "Spawn editor to capture an introductory message." ),
    ] = False
    prepend_prompt: __.typx.Annotated[
        bool,
        __.tyro.conf.arg(
            help = "Prepend mimeogram format instructions." ),
    ] = False
    recurse: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.tyro.conf.arg(
            aliases = ( '-r', '--recurse-directories', '--recursive' ),
            help = "Recurse into directories." ),
    ] = None
    strict: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.tyro.conf.arg(
            aliases = ( '--fail-on-invalid', ),
            help = "Fail on invalid contents? True, fail. False, skip." ),
    ] = None

    async def __call__( self, auxdata: __.Globals ) -> None:
        ''' Executes command to create mimeogram. '''
        await create( auxdata, self )

    def provide_configuration_edits( self ) -> __.DictionaryEdits:
        ''' Provides edits against configuration from options. '''
        edits: list[ __.DictionaryEdit ] = [ ]
        if None is not self.clip:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'create', 'to-clipboard' ), value = self.clip ) )
        if None is not self.recurse:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'acquire-parts', 'recurse-directories' ),
                value = self.recurse ) )
        if None is not self.strict:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'acquire-parts', 'fail-on-invalid' ),
                value = self.strict ) )
        return tuple( edits )


async def _acquire_prompt( auxdata: __.Globals ) -> str:
    from .prompt import acquire_prompt
    return await acquire_prompt( auxdata )


async def _copy_to_clipboard( mimeogram: str ) -> None:
    from pyperclip import copy
    try: copy( mimeogram )
    except Exception as exc:
        _scribe.exception( "Could not copy mimeogram to clipboard." )
        raise SystemExit( 1 ) from exc
    _scribe.info( "Copied mimeogram to clipboard." )


async def _edit_message( ) -> str:
    from .edit import edit_content
    try: return edit_content( )
    except Exception as exc:
        _scribe.exception( "Could not acquire user message." )
        raise SystemExit( 1 ) from exc


async def create( # pylint: disable=too-complex,too-many-locals,too-many-statements
    auxdata: __.Globals,
    command: Command, *,
    editor: __.cabc.Callable[
        [ ], __.cabc.Coroutine[ None, None, str ] ] = _edit_message,
    clipcopier: __.cabc.Callable[
        [ str ], __.cabc.Coroutine[ None, None, None ] ] = _copy_to_clipboard,
    prompter: __.cabc.Callable[
        [ __.Globals ],
        __.cabc.Coroutine[ None, None, str ] ] = _acquire_prompt,
) -> __.typx.Never:
    ''' Creates mimeogram. '''
    from exceptiongroup import ExceptionGroup
    from .acquirers import acquire
    from .formatters import format_mimeogram
    if command.edit:
        try: message = await editor( )
        except Exception as exc:
            _scribe.exception( "Could not acquire user message." )
            raise SystemExit( 1 ) from exc
    else: message = None
    # TODO: Factor into '__.exceptions.report_eg_members'.
    try: parts = await acquire( auxdata, command.sources )
    except ExceptionGroup as excg: # pyright: ignore
        for exc in excg.exceptions: # pyright: ignore
            _scribe.error( # noqa: TRY400
                "Could not acquire mimeogram parts.",
                exc_info = exc ) # pyright: ignore
        raise SystemExit( 1 ) from None
    except Exception as exc:
        _scribe.exception( "Could not acquire mimeogram parts." )
        raise SystemExit( 1 ) from exc
    mimeogram = format_mimeogram( parts, message = message )
    # TODO? Pass prompt to 'format_mimeogram'.
    if command.prepend_prompt:
        prompt = await prompter( auxdata )
        mimeogram = f"{prompt}\n\n{mimeogram}"
    options = auxdata.configuration.get( 'create', { } )
    if options.get( 'to-clipboard', False ):
        try: await clipcopier( mimeogram )
        except Exception as exc:
            _scribe.exception( "Could not copy mimeogram to clipboard." )
            raise SystemExit( 1 ) from exc
    else: print( mimeogram ) # TODO? Use output stream from configuration.
    raise SystemExit( 0 )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Sensitive filesystem locations relative to user homes. '''


from __future__ import annotations

from . import __


_scribe = __.produce_scribe( __name__ )


def discover_sensitive_locations( ) -> frozenset[ str ]:
    ''' Discovers sensitive locations relative to user home. '''
    return frozenset( (
        # Credentials and Keys
        '.ssh', '.aws', '.gnupg', '.gpg',
        # Cloud Services
        '.config/gcloud', '.azure', '.kube',
        '.terraform.d', '.chef',
        # Package Managers
        '.npm', '.pip', '.cargo', '.gem',
        '.gradle', '.m2', '.ivy2',
        # Browser Data
        '.mozilla', '.chrome', '.config/chromium',
        '.netscape', '.opera',
        # Database
        '.postgresql', '.mysql', '.redis',
        # Shell History and Config
        '.bash_history', '.zsh_history',
        '.bashrc', '.zshrc', '.profile',
        # Cryptocurrency
        '.bitcoin', '.ethereum',
        # Other Sensitive
        '.password-store', '.secrets', '.keys',
        '.config', '.local/share',
    ) )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common I/O primitives. '''


from . import imports as __
from . import asyncf as _asyncf


async def acquire_text_file_async(
    file: str | __.Path,
    charset: str = 'utf-8',
    deserializer: __.Absential[
        __.typx.Callable[ [ str ], __.typx.Any ] ] = __.absent,
) -> __.typx.Any:
    ''' Reads file asynchronously. '''
    from aiofiles import open as open_
    async with open_( file, encoding = charset ) as stream:
        data = await stream.read( )
    if not __.is_absent( deserializer ):
        return deserializer( data )
    return data


async def acquire_text_files_async(
    *files: str | __.Path,
    charset: str = 'utf-8',
    deserializer: __.Absential[
        __.typx.Callable[ [ str ], __.typx.Any ] ] = __.absent,
    return_exceptions: bool = False
) -> __.typx.Sequence[ __.typx.Any ]:
    ''' Reads files in parallel asynchronously. '''
    # TODO? Batch to prevent fd exhaustion over large file sets.
    return await _asyncf.gather_async(
        *(  acquire_text_file_async(
                file, charset = charset, deserializer = deserializer )
            for file in files ),
        error_message = 'Failure to read files.',
        return_exceptions = return_exceptions )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common constants, imports, and utilities. '''


# Expose everything from internal modules.
from .application import Information as ApplicationInformation
from .asyncf import *
from .dictedits import (
    Edit as                 DictionaryEdit,
    Edits as                DictionaryEdits,
    ElementsEntryEdit as    ElementsEntryDictionaryEdit,
    SimpleEdit as           SimpleDictionaryEdit,
)
from .distribution import Information as DistributionInformation
from .exceptions import *
from .generics import *
from .imports import *
from .inscription import (
    Control as InscriptionControl, Modes as InscriptionModes )
from .io import *
from .preparation import *
from .state import Globals



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Fundamental configuration. '''


from . import imports as __
from . import dictedits as _dictedits
from . import distribution as _distribution
from . import exceptions as _exceptions
from . import io as _io


class EnablementTristate( __.enum.Enum ): # TODO: Python 3.11: StrEnum
    ''' Disable, enable, or retain the natural state? '''

    Disable = 'disable'
    Retain = 'retain'
    Enable = 'enable'

    def __bool__( self ) -> bool:
        if self.Disable is self: return False
        if self.Enable is self: return True
        raise _exceptions.OperationInvalidity(
            'inert enablement tristate', 'boolean translation' )

    def is_retain( self ) -> bool:
        ''' Does enum indicate a retain state? '''
        return self.Retain is self


async def acquire(
    application_name: str,
    directories: __.PlatformDirs,
    distribution: _distribution.Information,
    edits: _dictedits.Edits = ( ),
    file: __.Absential[ __.Path ] = __.absent,
) -> __.AccretiveDictionary[ str, __.typx.Any ]:
    ''' Loads configuration as dictionary. '''
    if __.is_absent( file ):
        file = _discover_copy_template( directories, distribution )
    configuration = await _acquire( file )
    includes = await _acquire_includes(
        application_name, directories, configuration.get( 'includes', ( ) ) )
    for include in includes: configuration.update( include )
    for edit in edits: edit( configuration )
    return __.AccretiveDictionary( configuration )


async def _acquire( file: __.Path ) -> __.NominativeDictionary:
    from aiofiles import open as open_
    from tomli import loads
    async with open_( file ) as stream:
        return loads( await stream.read( ) )


async def _acquire_includes(
    application_name: str,
    directories: __.PlatformDirs,
    specs: tuple[ str, ... ]
) -> __.cabc.Sequence[ dict[ str, __.typx.Any ] ]:
    from itertools import chain
    from tomli import loads
    locations = tuple(
        __.Path( spec.format(
            user_configuration = directories.user_config_path,
            user_home = __.Path.home( ),
            application_name = application_name ) )
        for spec in specs )
    iterables = tuple(
        ( location.glob( '*.toml' ) if location.is_dir( ) else ( location, ) )
        for location in locations )
    return await _io.acquire_text_files_async(
        *( file for file in chain.from_iterable( iterables ) ),
        deserializer = loads )


def _discover_copy_template(
    directories: __.PlatformDirs, distribution: _distribution.Information
) -> __.Path:
    from shutil import copyfile
    file = directories.user_config_path / 'general.toml'
    if not file.exists( ):
        copyfile(
            distribution.provide_data_location(
                'configuration', 'general.toml' ), file )
    return file



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Generic types. '''
# TODO: Independent package.


from __future__ import annotations

from . import imports as __


T = __.typx.TypeVar( 'T' ) # generic
U = __.typx.TypeVar( 'U' ) # generic
E = __.typx.TypeVar( 'E', bound = Exception ) # error


class Result( __.ImmutableObject, __.typx.Generic[ T, E ] ):
    ''' Either a value or an error. '''
    # TODO: Protocol class.

    def is_error( self ) -> bool:
        ''' Returns ``True`` if error result. Else ``False``. '''
        return isinstance( self, Error )

    def is_value( self ) -> bool:
        ''' Returns ``True`` if value result. Else ``False``. '''
        return isinstance( self, Value )

    @__.abc.abstractmethod
    def extract( self ) -> T:
        ''' Extracts value from result. Else, raises error from result.

            Similar to Result.unwrap in Rust.
        '''
        raise NotImplementedError

    @__.abc.abstractmethod
    def transform(
        self, function: __.typx.Callable[ [ T ], U ]
    ) -> __.typx.Self | Result[ U, E ]:
        ''' Transforms value in value result. Ignores error result.

            Similar to Result.map in Rust.
        '''
        raise NotImplementedError


class Value( Result[ T, E ] ):
    ''' Result of successful computation. '''

    __match_args__ = ( 'value', )
    __slots__ = ( 'value', )

    value: T

    def __init__( self, value: T ): self.value = value

    def extract( self ) -> T: return self.value

    def transform(
        self, function: __.typx.Callable[ [ T ], U ]
    ) -> Result[ U, E ]: return Value( function( self.value ) )


class Error( Result[ T, E ] ):
    ''' Result of failed computation. '''

    __match_args__ = ( 'error', )
    __slots__ = ( 'error', )

    error: E

    def __init__( self, error: E ): self.error = error

    def extract( self ) -> __.typx.Never: raise self.error

    def transform(
        self, function: __.typx.Callable[ [ T ], U ]
    ) -> __.typx.Self: return self


GenericResult: __.typx.TypeAlias = Result[ __.typx.Any, Exception ]



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common imports and type aliases used throughout the package. '''

# pylint: disable=unused-import
# ruff: noqa: F401


from __future__ import annotations

import abc
import asyncio
import collections.abc as cabc
import enum
import os
import re
import sys
import types

from contextlib import AsyncExitStack as ExitsAsync
from dataclasses import (
    dataclass,
    field as dataclass_declare,
)
from logging import getLogger as produce_scribe
from pathlib import Path
from uuid import uuid4

import typing_extensions as typx
import tyro

from absence import Absential, absent, is_absent
from accretive.qaliases import AccretiveDictionary
from frigid.qaliases import (
    ImmutableClass,
    ImmutableDictionary,
    ImmutableObject,
    ImmutableProtocolClass,
    reclassify_modules_as_immutable,
)
from platformdirs import PlatformDirs


ComparisonResult: typx.TypeAlias = bool | types.NotImplementedType
NominativeDictionary: typx.TypeAlias = cabc.MutableMapping[ str, typx.Any ]
ImmutableNominativeDictionary: typx.TypeAlias = cabc.Mapping[ str, typx.Any ]


@typx.dataclass_transform( frozen_default = True, kw_only_default = True )
class ImmutableStandardDataclass( ImmutableClass ):
    ''' Metaclass for immutable standard dataclasses. (Typechecker hack.) '''


@typx.dataclass_transform( frozen_default = True, kw_only_default = True )
class ImmutableStandardProtocolDataclass( ImmutableProtocolClass ):
    ''' Metaclass for immutable standard dataclasses. (Typechecker hack.) '''


package_name = __name__.split( '.', maxsplit = 1 )[ 0 ]
simple_tyro_class = tyro.conf.configure( )
standard_dataclass = dataclass( frozen = True, kw_only = True, slots = True )
standard_tyro_class = tyro.conf.configure( tyro.conf.OmitArgPrefixes )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Helper functions for async execution. '''


from __future__ import annotations

from . import imports as __
from . import exceptions as _exceptions
from . import generics as _generics


async def gather_async(
    *operands: __.typx.Any,
    return_exceptions: __.typx.Annotated[
        bool,
        __.typx.Doc( ''' Raw or wrapped results. Wrapped, if true. ''' )
    ] = False,
    error_message: str = 'Failure of async operations.',
    ignore_nonawaitables: __.typx.Annotated[
        bool,
        __.typx.Doc(
            ''' Ignore or error on non-awaitables. Ignore, if true. ''' )
    ] = False,
) -> tuple[ __.typx.Any, ... ]:
    ''' Gathers results from invocables concurrently and asynchronously. '''
    # TODO: Overload signature for 'return_exceptions'.
    from exceptiongroup import ExceptionGroup # TODO: Python 3.11: builtin
    if ignore_nonawaitables:
        results = await _gather_async_permissive( operands )
    else:
        results = await _gather_async_strict( operands )
    if return_exceptions: return tuple( results )
    errors = tuple( result.error for result in results if result.is_error( ) )
    if errors: raise ExceptionGroup( error_message, errors )
    return tuple( result.extract( ) for result in results )


async def intercept_error_async(
    awaitable: __.cabc.Awaitable[ __.typx.Any ]
) -> _generics.Result[ object, Exception ]:
    ''' Converts unwinding exceptions to error results.

        Exceptions, which are not instances of :py:exc:`Exception` or one of
        its subclasses, are allowed to propagate. In particular,
        :py:exc:`KeyboardInterrupt` and :py:exc:`SystemExit` must be allowed
        to propagate to be consistent with :py:class:`asyncio.TaskGroup`
        behavior.

        Helpful when working with :py:func:`asyncio.gather`, for example,
        because exceptions can be distinguished from computed values
        and collected together into an exception group.

        In general, it is a bad idea to swallow exceptions. In this case,
        the intent is to add them into an exception group for continued
        propagation.
    '''
    try: return _generics.Value( await awaitable )
    except Exception as exc: # pylint: disable=broad-exception-caught
        return _generics.Error( exc )


async def _gather_async_permissive(
    operands: __.cabc.Sequence[ __.typx.Any ]
) -> __.cabc.Sequence[ __.typx.Any ]:
    from asyncio import gather # TODO? Python 3.11: TaskGroup
    awaitables: dict[ int, __.cabc.Awaitable[ __.typx.Any ] ] = { }
    results: list[ _generics.GenericResult ] = [ ]
    for i, operand in enumerate( operands ):
        if isinstance( operand, __.cabc.Awaitable ):
            awaitables[ i ] = (
                intercept_error_async( __.typx.cast(
                    __.cabc.Awaitable[ __.typx.Any ], operand ) ) )
            results.append( _generics.Value( None ) )
        else: results.append( _generics.Value( operand ) )
    results_ = await gather( *awaitables.values( ) )
    for i, result in zip( awaitables.keys( ), results_ ):
        results[ i ] = result
    return results


async def _gather_async_strict(
    operands: __.cabc.Sequence[ __.typx.Any ]
) -> __.cabc.Sequence[ __.typx.Any ]:
    from inspect import isawaitable, iscoroutine
    from asyncio import gather # TODO? Python 3.11: TaskGroup
    awaitables: list[ __.cabc.Awaitable[ __.typx.Any ] ] = [ ]
    for operand in operands: # Sanity check.
        if isawaitable( operand ): continue
        for operand_ in operands: # Cleanup.
            if iscoroutine( operand_ ): operand_.close( )
        raise _exceptions.AsyncAssertionFailure( operand )
    for operand in operands:
        awaitables.append( intercept_error_async( __.typx.cast(
            __.cabc.Awaitable[ __.typx.Any ], operand ) ) )
    return await gather( *awaitables )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Parsers for mimeograms and their constituents. '''


from __future__ import annotations

from . import __
from . import parts as _parts


_scribe = __.produce_scribe( __name__ )


def parse( mgtext: str ) -> __.cabc.Sequence[ _parts.Part ]:
    ''' Parses mimeogram. '''
    # TODO? Accept 'strict' flag.
    from .exceptions import MimeogramParseFailure
    if not mgtext.strip( ):
        raise MimeogramParseFailure( reason = "Empty mimeogram." )
    boundary = _extract_boundary( mgtext )
    ptexts = _separate_parts( mgtext, boundary )
    parts: list[ _parts.Part ] = [ ]
    for i, ptext in enumerate( ptexts, 1 ):
        try: part = parse_part( ptext )
        except MimeogramParseFailure:
            _scribe.exception( f"Parse failure on part {i}." )
            continue
        parts.append( part )
        _scribe.debug( f"Parsed part {i} with location '{part.location}'." )
    _scribe.debug( "Parsed {} parts.".format( len( parts ) ) )
    return parts


def parse_part( ptext: str ) -> _parts.Part:
    ''' Parses mimeogram part. '''
    descriptor, content = _parse_descriptor_and_content( ptext )
    _validate_descriptor( descriptor )
    mimetype, charset, linesep = (
        _parse_mimetype( descriptor[ 'Content-Type' ] ) )
    return _parts.Part(
        location = descriptor[ 'Content-Location' ],
        mimetype = mimetype, charset = charset, linesep = linesep,
        content = content )


_BOUNDARY_REGEX = __.re.compile(
    r'''^--====MIMEOGRAM_[0-9a-fA-F]{16,}====\s*$''',
    __.re.IGNORECASE | __.re.MULTILINE )
def _extract_boundary( content: str ) -> str:
    ''' Extracts first mimeogram boundary. '''
    mobject = _BOUNDARY_REGEX.search( content )
    if mobject:
        boundary = mobject.group( )
        # Windows clipboard has CRLF newlines. Strip CR before display.
        boundary_s = boundary.rstrip( '\r' )
        _scribe.debug( f"Found boundary: {boundary_s}" )
        # Return with trailing newline to ensure parts are properly split.
        return f"{boundary}\n"
    from .exceptions import MimeogramParseFailure
    raise MimeogramParseFailure( reason = "No mimeogram boundary found." )


_DESCRIPTOR_REGEX = __.re.compile(
    r'''^(?P<name>[\w\-]+)\s*:\s*(?P<value>.*)$''' )
def _parse_descriptor_and_content(
    content: str
) -> tuple[ __.cabc.Mapping[ str, str ], str ]:
    descriptor: __.cabc.Mapping[ str, str ] = { }
    lines: list[ str ] = [ ]
    in_matter = False
    for line in content.splitlines( ):
        if in_matter:
            lines.append( line )
            continue
        line_s = line.strip( )
        if not line_s:
            in_matter = True
            continue
        mobject = _DESCRIPTOR_REGEX.fullmatch( line_s )
        if not mobject:
            _scribe.warning( "No blank line after headers." )
            in_matter = True
            lines.append( line )
            continue
        name = '-'.join( map(
            str.capitalize, mobject.group( 'name' ).split( '-' ) ) )
        value = mobject.group( 'value' )
        # TODO: Detect duplicates.
        descriptor[ name ] = value
    _scribe.debug( f"Descriptor: {descriptor}" )
    return descriptor, '\n'.join( lines )


_QUOTES = '"\''
def _parse_mimetype( header: str ) -> tuple[ str, str, _parts.LineSeparators ]:
    ''' Extracts MIME type and charset from Content-Type header. '''
    parts = [ p.strip( ) for p in header.split( ';' ) ]
    mimetype = parts[ 0 ]
    charset = 'utf-8'
    linesep = _parts.LineSeparators.LF
    for part in parts[ 1: ]:
        if part.startswith( 'charset=' ):
            charset = part[ 8: ].strip( _QUOTES )
        if part.startswith( 'linesep=' ):
            linesep = _parts.LineSeparators[
                part[ 8: ].strip( _QUOTES ).upper( ) ]
    return mimetype, charset, linesep


def _separate_parts( content: str, boundary: str ) -> list[ str ]:
    ''' Splits content into parts using boundary. '''
    boundary_s = boundary.rstrip( )
    final_boundary = f"{boundary_s}--"
    # Detect final boundary and trailing text first.
    final_parts = content.split( final_boundary )
    if len( final_parts ) > 1:
        _scribe.debug( "Found final boundary." )
        content_with_parts = final_parts[ 0 ]
        trailing_text = final_parts[ 1 ].strip( )
        if trailing_text: _scribe.debug( "Found trailing text." )
    else:
        _scribe.warning( "No final boundary found." )
        content_with_parts = content
    # Split remaining content on regular boundary and skip leading text.
    parts = content_with_parts.split( boundary )[ 1: ]
    _scribe.debug( "Found {} parts to parse.".format( len( parts ) ) )
    return parts


_DESCRIPTOR_INDICES_REQUISITE = frozenset( (
    'Content-Location', 'Content-Type' ) )
def _validate_descriptor(
    descriptor: __.cabc.Mapping[ str, str ]
) -> __.cabc.Mapping[ str, str ]:
    from .exceptions import MimeogramParseFailure
    names = _DESCRIPTOR_INDICES_REQUISITE - descriptor.keys( )
    if names:
        reason = (
            "Missing required headers: {awol}".format(
                awol = ', '.join( names ) ) )
        _scribe.warning( reason )
        raise MimeogramParseFailure( reason = reason )
    return descriptor # TODO: Return immutable.



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' File content updates. '''


from __future__ import annotations

from . import __
from . import fsprotect as _fsprotect
from . import interactions as _interactions
from . import interfaces as _interfaces
from . import parts as _parts


_scribe = __.produce_scribe( __name__ )


class ReviewModes( __.enum.Enum ): # TODO: Python 3.11: StrEnum
    ''' Controls how updates are reviewed and applied. '''

    Silent =    'silent'        # Apply parts without review.
    # Aggregate = 'aggregate'     # Git-style aggregated diff for all parts
    Partitive = 'partitive'     # Interactively review each part.


# pylint: disable=bad-reversed-sequence,unsubscriptable-object
# pylint: disable=unsupported-assignment-operation,unsupported-membership-test
class Reverter(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Backup and restore filesystem state. '''

    originals: dict[ __.Path, str ] = (
        __.dataclass_declare( default_factory = dict ) )
    revisions: list[ __.Path ] = (
        __.dataclass_declare( default_factory = list ) )

    async def save( self, part: _parts.Part, path: __.Path ) -> None:
        ''' Saves original file content if it exists. '''
        from .exceptions import ContentAcquireFailure
        if not path.exists( ): return
        try:
            content = (
                await __.acquire_text_file_async(
                    path, charset = part.charset ) )
        except Exception as exc: raise ContentAcquireFailure( path ) from exc
        self.originals[ path ] = content

    async def restore( self ) -> None:
        ''' Restores files to original contents in reverse order. '''
        # TODO: async parallel fanout
        from .exceptions import ContentUpdateFailure
        for path in reversed( self.revisions ):
            if path in self.originals:
                try:
                    await _update_content_atomic(
                        path, self.originals[ path ] )
                except ContentUpdateFailure:
                    _scribe.exception( "Failed to restore {path}" )
            else: path.unlink( )
# pylint: enable=bad-reversed-sequence,unsubscriptable-object
# pylint: enable=unsupported-assignment-operation,unsupported-membership-test


# pylint: disable=no-member,not-an-iterable
class Queue(
    metaclass = __.ImmutableStandardDataclass,
    decorators = ( __.standard_dataclass, ),
):
    ''' Manages queued file updates for batch application. '''

    updates: list[ tuple[ _parts.Part, __.Path, str ] ] = (
        __.dataclass_declare( default_factory = list ) )
    reverter: Reverter = (
        __.dataclass_declare( default_factory = Reverter ) )

    def enqueue(
        self, part: _parts.Part, target: __.Path, content: str
    ) -> None:
        ''' Adds a file update to queue. '''
        self.updates.append( ( part, target, content ) )

    async def apply( self ) -> None:
        ''' Applies all queued updates with parallel async fanout. '''
        try:
            await __.gather_async(
                *(  self.reverter.save( part, target )
                    for part, target, _ in self.updates ),
                error_message = "Failed to backup files." )
            await __.gather_async(
                *(  _update_content_atomic(
                        target, content, charset = part.charset )
                    for part, target, content in self.updates ),
                error_message = "Failed to apply updates." )
        except Exception:
            await self.reverter.restore( )
            raise
        for _, target, _ in self.updates:
            self.reverter.revisions.append( target )
# pylint: enable=no-member,not-an-iterable


async def update( # pylint: disable=too-many-arguments,too-many-locals
    auxdata: __.Globals,
    parts: __.cabc.Sequence[ _parts.Part ],
    mode: ReviewModes,
    base: __.Absential[ __.Path ] = __.absent,
    interactor: __.Absential[ _interfaces.PartInteractor ] = __.absent,
    protector: __.Absential[ _fsprotect.Protector ] = __.absent,
) -> None:
    ''' Updates filesystem locations from mimeogram. '''
    if __.is_absent( base ): base = __.Path( )
    if __.is_absent( protector ):
        protector = _fsprotect.Cache.from_configuration( auxdata = auxdata )
    queue = Queue( )
    for part in parts:
        if part.location.startswith( 'mimeogram://' ): continue
        destination = _derive_location( part.location, base = base )
        target = _parts.Target(
            part = part,
            destination = destination,
            protection = protector.verify( destination ) )
        action, content = await update_part(
            auxdata, target, mode = mode, interactor = interactor )
        if _parts.Resolutions.Ignore is action: continue
        queue.enqueue( target.part, target.destination, content )
    await queue.apply( )


async def update_part(
    auxdata: __.Globals,
    target: _parts.Target,
    mode: ReviewModes,
    interactor: __.Absential[ _interfaces.PartInteractor ] = __.absent,
) -> tuple[ _parts.Resolutions, str ]:
    ''' Updates filesystem location from mimeogram part. '''
    content = target.part.content
    if ReviewModes.Partitive is mode:
        return await _interactions.interact( target, interactor = interactor )
    options = auxdata.configuration.get( 'update-parts', { } )
    if target.protection and not options.get( 'disable-protections', False ):
        _scribe.warning(
            f"Skipping protected path: {target.destination} "
            f"Reason: {target.protection.description}" )
        return _parts.Resolutions.Ignore, content
    return _parts.Resolutions.Apply, content


def _derive_location(
    location: __.typx.Annotated[
        str, __.typx.Doc( "Part location (URL or filesystem path)." ) ],
    base: __.typx.Annotated[
        __.Absential[ __.Path ],
        __.typx.Doc(
            "Base path for relative locations. "
            "Defaults to current directory." )
    ] = __.absent,
) -> __.Path:
    ''' Resolves part location to filesystem path. '''
    import os.path as ospath
    from urllib.parse import urlparse
    from .exceptions import LocationInvalidity
    try: url = urlparse( location )
    except Exception as exc: raise LocationInvalidity( location ) from exc
    path = __.Path( location )
    scheme = 'file' if path.drive else url.scheme
    match scheme:
        case '' | 'file': pass
        case _: raise LocationInvalidity( location )
    location_ = __.Path( ospath.expanduser( ospath.expandvars( url.path ) ) )
    if location_.is_absolute( ): return location_
    if not __.is_absent( base ): return ( base / location_ ).resolve( )
    return __.Path( ) / location_


async def _update_content_atomic(
    location: __.Path,
    content: str,
    charset: str = 'utf-8',
    linesep: _parts.LineSeparators = _parts.LineSeparators.LF
) -> None:
    ''' Updates file content atomically, if possible. '''
    import aiofiles.os as os # pylint: disable=consider-using-from-import
    from aiofiles.tempfile import NamedTemporaryFile
    location.parent.mkdir( parents = True, exist_ok = True )
    content = linesep.nativize( content )
    has_error = False
    async with NamedTemporaryFile(
        delete = False,
        dir = location.parent,
        suffix = f"{location.suffix}.tmp",
    ) as stream:
        filename = str( stream.name )
        try: await stream.write( content.encode( charset ) )
        except Exception: # pylint: disable=broad-exception-caught
            has_error = True
    # Windows: Replace must happen after file handle is closed.
    if not has_error:
        try: await os.replace( filename, str( location ) )
        except Exception: # pylint: disable=broad-exception-caught
            has_error = True
    if await os.path.exists( filename ):
        try: await os.remove( filename )
        except Exception: # pylint: disable=broad-exception-caught
            _scribe.warning( f"Could not remove temporary file: {filename}" )
    if has_error:
        from .exceptions import ContentUpdateFailure
        raise ContentUpdateFailure( location )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Formatting of mimeogram bundles. '''


from __future__ import annotations

from . import __
from . import parts as _parts


def format_mimeogram(
    parts: __.cabc.Sequence[ _parts.Part ],
    message: __.typx.Optional[ str ] = None,
) -> str:
    ''' Formats parts into mimeogram. '''
    if not parts and message is None:
        from .exceptions import MimeogramFormatEmpty
        raise MimeogramFormatEmpty( )
    boundary = "====MIMEOGRAM_{uuid}====".format( uuid = __.uuid4( ).hex )
    lines: list[ str ] = [ ]
    if message:
        message_part = _parts.Part(
            location = 'mimeogram://message',
            mimetype = 'text/plain', # TODO? Markdown
            charset = 'utf-8',
            linesep = _parts.LineSeparators.LF,
            content = message )
        lines.append( format_part( message_part, boundary ) )
    for part in parts:
        lines.append( format_part( part, boundary ) )
    lines.append( f"--{boundary}--" )
    return '\n'.join( lines )


def format_part( part: _parts.Part, boundary: str ) -> str:
    ''' Formats part with boundary marker and headers. '''
    return '\n'.join( (
        f"--{boundary}",
        f"Content-Location: {part.location}",
        f"Content-Type: {part.mimetype}; "
        f"charset={part.charset}; "
        f"linesep={part.linesep.name}",
        '',
        part.content ) )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' System pager interaction. '''


from __future__ import annotations

from . import __


_scribe = __.produce_scribe( __name__ )


def discover_pager( ) -> __.cabc.Callable[ [ str ], None ]:
    ''' Discovers pager and returns executor function. '''
    from shutil import which
    from subprocess import run # nosec B404
    pager = __.os.environ.get( 'PAGER', 'less' )
    for pager_ in ( pager, 'less', 'more' ):
        if ( pager := which( pager_ ) ):
            match __.sys.platform:
                case 'win32':
                    # Windows 'more.com' does not support UTF-8.
                    if pager.lower( ).endswith( '\\more.com' ): continue
                case _: pass
            break
    else: pager = ''

    if pager:

        # TODO? async
        def pager_executor( filename: str ) -> None:
            ''' Executes pager with file. '''
            run( ( pager, filename ), check = True ) # nosec B603

        return pager_executor

    # TODO? async
    def console_display( filename: str ) -> None:
        ''' Prints file to stdout and waits for ENTER key. '''
        with open( filename, 'r', encoding = 'utf-8' ) as stream:
            content = stream.read( )
        print( f"\n\n{content}\n\n" )
        if __.sys.stdin.isatty( ): input( "Press Enter to continue..." )

    _scribe.warning( "Could not find pager program for display." )
    return console_display


def display_content(
    content: str, *,
    suffix: str = '.txt',
    pager_discoverer: __.cabc.Callable[
        [ ], __.cabc.Callable[ [ str ], None ] ] = discover_pager,
) -> None:
    ''' Displays content via discovered pager. '''
    from .exceptions import PagerFailure
    pager = pager_discoverer( )
    import tempfile
    from pathlib import Path
    # Using delete = False to handle file cleanup manually. This ensures
    # the file handle is properly closed before the pager attempts to read it,
    # which is particularly important on Windows where open files cannot be
    # simultaneously accessed by other processes without a read share.
    with tempfile.NamedTemporaryFile(
        mode = 'w', suffix = suffix, delete = False, encoding = 'utf-8'
    ) as tmp:
        filename = tmp.name
        tmp.write( content )
    try: pager( filename )
    except Exception as exc: raise PagerFailure( cause = exc ) from exc
    finally:
        try: Path( filename ).unlink( )
        except Exception: # pylint: disable=broad-exception-caught
            _scribe.exception( f"Failed to cleanup {filename}" )
